/*
 * Simple heap priority queue
 *
 * Grigoriy Sitkarev, <sitkarev@unixkomi.ru>
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "common.h"
#include "macros.h"
#include "xmalloc.h"
#include "failures.h"
#include "prioq.h"

#define LEFT_CHILD(i)	(((i) * 2) + 1)
#define RIGHT_CHILD(i)	(((i) * 2) + 2)
#define PARENT(i)	(((i-1)/2))

struct prioq {
	size_t			used;
	size_t			size;
	struct prioq_head	**heap;
	prioqcompare_t		compare;
};

int prioq_put(struct prioq *queue, struct prioq_head *h)
{
	int i;

	if (!queue || !h)
		xerror(1, "prioq_put: NULL args");

	if (queue->used == queue->size) {
		queue->heap = xrealloc0(queue->heap,
				sizeof(struct pioq_head *) * queue->size,
				sizeof(struct prioq_head *) * (queue->size + 32));
		queue->size += 32;
	}
	i = queue->used++;
	while (i && queue->compare(queue->heap[PARENT(i)], h) < 0) {
		queue->heap[PARENT(i)]->index = i;
		queue->heap[i] = queue->heap[PARENT(i)];
		i = PARENT(i);
	}
	queue->heap[i] = h;
	h->index = i;
	return 0;
}

static __inline__ void prioq_heapify(struct prioq *q, int index)
{
	int largest;
	struct prioq_head *tmp;

	for (index = 0, largest = LEFT_CHILD(index); largest < q->used;
	     index = largest, largest = LEFT_CHILD(index)) {
		if (q->compare(q->heap[largest], q->heap[largest+1]) < 0)
			largest++;
		if (q->compare(q->heap[index], q->heap[largest]) < 0) {
			tmp = q->heap[largest];
			q->heap[largest] = q->heap[index];
			q->heap[largest]->index = largest;
			q->heap[index] = tmp;
			q->heap[index]->index = index;
		} else
			return;
	}
	if (largest == q->used &&
	    q->compare(q->heap[index], q->heap[largest]) < 0) {
		tmp = q->heap[largest];
		q->heap[largest] = q->heap[index];
		q->heap[largest]->index = largest;
		q->heap[index] = tmp;
		q->heap[index]->index = index;
	}
}

struct prioq_head *prioq_peek(struct prioq *queue)
{
	return queue->used ? queue->heap[0] : NULL;
}

struct prioq_head *prioq_get(struct prioq *queue)
{
	struct prioq_head *h;

	if (!queue)
		xerror(1, "prioq_get: queue NULL");

	if (queue->used == 0)
		return NULL;
	h = queue->heap[0];
	h->index = -1; /* mark as deleted */
	queue->used--;
	if (queue->used) {
		queue->heap[0] = queue->heap[queue->used];
		queue->heap[0]->index = 0;
		prioq_heapify(queue, 0);
	}
	return h;
}

void prioq_remove(struct prioq *q, struct prioq_head *h)
{
	int i;

	if (!q || !h)
		xerror(1, "prioq_remove: args NULL");

	if (h->index < 0 || h->index > q->used)
		xerror(1, "prioq_remove: queue=%p head=%p invalid index=%d",
				q, h, h->index);
	if (q->heap[h->index] != h)
		xerror(1, "prioq_remove: %p not in queue=%p idx=%d",
				h, q, h->index);
	for (i = h->index; i > 0; i = PARENT(i)) {
		q->heap[i] = q->heap[PARENT(i)];
		q->heap[i]->index = i;
	}
	q->used--;
	h->index = -1;	/* mark as deleted */
	if (q->used) {
		q->heap[0] = q->heap[q->used];
		q->heap[0]->index = 0;
		prioq_heapify(q, 0);
	}
}

struct prioq *prioq_new(prioqcompare_t compare)
{
	struct prioq *queue;

	if (!compare)
		xerror(1, "prioq_new: compare NULL");

	queue = xmalloc0(sizeof(*queue));
	queue->size = 32;
	queue->used = 0;
	queue->compare = compare;
	queue->heap = xmalloc0(sizeof(struct prioq_head *) * queue->size);
	return queue;
}

void prioq_free(struct prioq **queue)
{
	if (!queue || *queue == NULL)
		xerror(1, "prioq_free: queue NULL");

	if ((*queue)->used)
		xwarning("prioq_free: queue %p not empty", *queue);

	xfree((*queue)->heap);
	xfree(*queue);
	*queue = NULL;
}
